import 'dart:ui';

import 'package:arrow_path/arrow_path.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

class PumaGuideItem {
  final String? title;
  final String? subtitle;
  final bool debugShowContentContainer;
  final String? description;
  final GlobalKey? globalKey;
  final ShapeFocus? shapeFocus;
  final double contentWidth;
  final double contentHeight;
  final bool showSkip;
  final bool showNext;
  final bool hapticFeedback;
  final bool invertContentIfNotFitBellow;
  final String nextWording;
  final String finishWording;
  final String skipWording;
  final double distanceFromObject;
  final bool invertContentOrder;
  final Color overlayBackgroundColor;
  final Color? titleBackgroundColor;
  final Widget Function()? builder;
  final Path Function(Path path, double contentWidth, double contentHeight,
      bool fitBellow, double distanceFromObject)? arrowBuilder;
  final Widget Function({Function skip, Function next})? controlAreaBuilder;

  PumaGuideItem({
    this.title,
    this.subtitle,
    this.description,
    this.globalKey,
    this.shapeFocus,
    this.debugShowContentContainer = false,
    this.invertContentIfNotFitBellow = true,
    this.distanceFromObject = 80,
    this.invertContentOrder = false,
    this.hapticFeedback = true,
    this.contentWidth = 280,
    this.contentHeight = 280,
    this.showSkip = true,
    this.showNext = true,
    this.overlayBackgroundColor = Colors.black,
    this.arrowBuilder,
    this.titleBackgroundColor,
    this.builder,
    this.controlAreaBuilder,
    this.nextWording = "Next",
    this.finishWording = "Finish",
    this.skipWording = "Skip",
  });
}

class PumaGuide {
  static start(BuildContext context, List<PumaGuideItem> children,
      Function dismissCallback) async {
    int count = 0;
    var size = MediaQuery.of(context).size;
    OverlayState? overlayState = Overlay.of(context);
    List<OverlayEntry> entrys = [];
    children.forEach((element) async {
      var offset = _capturePositionWidget(element.globalKey);
      var sizeWidget = _getSizeWidget(element.globalKey);
      entrys.add(
        OverlayEntry(
          builder: (context) {
            final next = () {
              if (element.hapticFeedback) HapticFeedback.mediumImpact();
              entrys[count].remove();
              if (entrys.isEmpty) {
                dismissCallback();
              }
              count++;
              if (count != entrys.length) {
                overlayState?.insert(entrys[count]);
              }
            };
            final skip = () {
              if (element.hapticFeedback) HapticFeedback.mediumImpact();
              entrys[count].remove();
              dismissCallback();
            };

            List<Widget> content = [
              if (element.builder != null) element.builder!(),
              if (element.builder == null)
                ...defaultBuilder(context, element.title, element.subtitle,
                    element.description, element.titleBackgroundColor),
              if (element.controlAreaBuilder != null)
                element.controlAreaBuilder!(skip: skip, next: next),
              if (element.controlAreaBuilder == null &&
                  (element.showSkip || element.showNext))
                defaultControlButtons(
                    element.showSkip ? skip : null,
                    element.showNext ? next : null,
                    count,
                    children.length,
                    element.finishWording,
                    element.nextWording,
                    element.skipWording)
            ];

            final screenHeight = MediaQuery.of(context).size.height -
                MediaQuery.of(context).padding.bottom -
                MediaQuery.of(context).padding.top;

            final yOffset =
                offset.dy + sizeWidget.height + element.distanceFromObject;

            final fitBellow = screenHeight - yOffset > element.contentHeight;

            return GestureDetector(
              onTap: next,
              child: Scaffold(
                backgroundColor: Colors.transparent,
                body: Stack(
                  children: [
                    CustomPaint(
                      size: size,
                      painter: HolePainter(
                          shapeFocus: element.shapeFocus,
                          dx: offset.dx + (sizeWidget.width / 2),
                          dy: offset.dy + (sizeWidget.height / 2),
                          width: sizeWidget.width,
                          overlayBackgroundColor:
                              element.overlayBackgroundColor,
                          height: sizeWidget.height),
                    ),
                    CustomPaint(
                      size: size,
                      painter: ArrowPainter(
                          context: context,
                          distanceFromObject: element.distanceFromObject,
                          dx: offset.dx + (sizeWidget.width / 2),
                          dy: offset.dy + (sizeWidget.height / 2),
                          arrowBuilder: element.arrowBuilder,
                          contentHeight: element.contentHeight,
                          contentWidth: element.contentWidth,
                          width: sizeWidget.width,
                          height: sizeWidget.height,
                          fitBellow: fitBellow,
                          color: element.titleBackgroundColor ??
                              Theme.of(context).primaryColor),
                    ),
                    Positioned(
                      top: fitBellow ? yOffset : null,
                      bottom: !fitBellow
                          ? MediaQuery.of(context).size.height -
                              offset.dy +
                              element.distanceFromObject
                          : null,
                      left: 0,
                      right: 0,
                      child: Center(
                        child: Container(
                          color: element.debugShowContentContainer
                              ? Colors.red
                              : null,
                          width: element.contentWidth,
                          height: element.contentHeight,
                          child: Column(
                            mainAxisAlignment: fitBellow
                                ? MainAxisAlignment.start
                                : MainAxisAlignment.end,
                            mainAxisSize: MainAxisSize.min,
                            children: element.invertContentOrder ||
                                    (element.invertContentIfNotFitBellow &&
                                        !fitBellow)
                                ? content.reversed.toList()
                                : content,
                          ),
                        ),
                      ),
                    )
                  ],
                ),
              ),
            );
          },
        ),
      );
    });

    overlayState?.insert(entrys[0]);
  }

  static Offset _capturePositionWidget(GlobalKey? key) {
    RenderBox? renderPosition =
        key!.currentContext!.findRenderObject() as RenderBox;

    return renderPosition.localToGlobal(Offset.zero);
  }

  static Size _getSizeWidget(GlobalKey? key) {
    RenderBox? renderSize =
        key!.currentContext?.findRenderObject() as RenderBox;
    return renderSize.size;
  }
}

class ArrowPainter extends CustomPainter {
  final BuildContext context;
  final double distanceFromObject;
  final double? dx;
  final double? dy;
  final double? width;
  final double? height;
  final double contentWidth;
  final double contentHeight;
  final Color color;
  final Path Function(Path path, double contentWidth, double contentHeight,
      bool fitBellow, double distanceFromObject)? arrowBuilder;
  final bool fitBellow;

  ArrowPainter({
    required this.context,
    required this.distanceFromObject,
    this.dx,
    this.dy,
    this.arrowBuilder,
    this.width,
    this.height,
    required this.contentWidth,
    required this.contentHeight,
    required this.fitBellow,
    required this.color,
  });

  @override
  void paint(Canvas canvas, Size size) {
    Path path;
    Paint paint = Paint()
      ..color = Colors.black
      ..style = PaintingStyle.stroke
      ..strokeCap = StrokeCap.round
      ..strokeJoin = StrokeJoin.round
      ..strokeWidth = 3.0;

    path = Path();
    if (fitBellow) {
      path.moveTo(size.width / 2, dy! + height! / 2 + distanceFromObject);
    } else {
      path.moveTo(size.width / 2, (dy! - height! / 2 - distanceFromObject));
    }
    if (arrowBuilder != null) {
      path = arrowBuilder!(
          path, contentWidth, contentHeight, fitBellow, distanceFromObject);
    }
    path = ArrowPath.make(path: path);
    canvas.drawPath(path, paint..color = color);
  }

  @override
  bool shouldRepaint(CustomPainter oldDelegate) {
    return false;
  }
}

class HolePainter extends CustomPainter {
  final double? dx;
  final double? dy;
  final double? width;
  final double? height;
  final ShapeFocus? shapeFocus;
  final Color overlayBackgroundColor;

  HolePainter({
    this.dx,
    this.dy,
    this.width,
    this.height,
    this.shapeFocus = ShapeFocus.oval,
    required this.overlayBackgroundColor,
  });
  @override
  void paint(Canvas canvas, Size size) {
    final paint = Paint();
    paint.color = overlayBackgroundColor;

    if (shapeFocus == ShapeFocus.oval) {
      canvas.drawPath(
          Path.combine(
            PathOperation.difference,
            Path()..addRect(Rect.fromLTWH(0, 0, size.width, size.height)),
            Path()
              ..addOval(
                  Rect.fromCircle(center: Offset(dx!, dy!), radius: width!))
              ..close(),
          ),
          paint);
    } else {
      canvas.drawPath(
          Path.combine(
            PathOperation.difference,
            Path()..addRect(Rect.fromLTWH(0, 0, size.width, size.height)),
            Path()
              ..addRect(Rect.fromLTWH(
                  dx! - (width! / 2), dy! - (height! / 2), width!, height!))
              ..close(),
          ),
          paint);
    }
  }

  @override
  bool shouldRepaint(CustomPainter oldDelegate) {
    return false;
  }
}

Widget defaultControlButtons(Function? skip, Function? next, int index,
    int total, String finishWording, String nextWording, String skipWording) {
  return Wrap(
    spacing: 8.0,
    runSpacing: 8.0,
    children: [
      if (skip != null)
        (OutlinedButton(
          child: Text(
            index + 1 == total ? finishWording : skipWording,
            style: TextStyle(color: Colors.white),
          ),
          onPressed: () => skip(),
        )),
      if (next != null && index + 1 < total)
        (ElevatedButton(
          style: ButtonStyle(
              backgroundColor:
                  MaterialStateProperty.all(Colors.grey.withAlpha(150))),
          child: Text(
            "$nextWording (${index + 1}/$total)",
            style: TextStyle(color: Colors.white),
          ),
          onPressed: () => next(),
        )),
    ],
  );
}

List<Widget> defaultBuilder(BuildContext context, String? title,
    String? subtitle, String? description, Color? color) {
  final _color = color != null ? color : Theme.of(context).primaryColor;
  final List<Widget> content = [];

  if (title != null) {
    content.add(Container(
      color: _color,
      child: Padding(
        padding: const EdgeInsets.all(16),
        child: Text(
          title,
          textAlign: TextAlign.center,
          style: Theme.of(context).textTheme.button,
        ),
      ),
    ));
  }

  if (subtitle != null || description != null) {
    content.add(Padding(
      padding: const EdgeInsets.only(bottom: 16.0, top: 16),
      child: Column(
        mainAxisAlignment: MainAxisAlignment.center,
        crossAxisAlignment: CrossAxisAlignment.center,
        children: [
          if (subtitle != null)
            Text(
              subtitle,
              textAlign: TextAlign.center,
              style: Theme.of(context).textTheme.bodyText1,
            ),
          if (description != null)
            Text(
              description,
              textAlign: TextAlign.center,
              style: Theme.of(context).textTheme.bodyText2,
            ),
        ],
      ),
    ));
  }

  return content;
}

enum ShapeFocus { oval, square }

Path arrowPointingLeft(Path path, double contentWidth, double contentHeight,
    bool fitBellow, double distanceFromObject) {
  if (fitBellow) {
    path.relativeQuadraticBezierTo(30, -30, 0, -distanceFromObject + 20);
  } else {
    path.relativeQuadraticBezierTo(30, 30, 0, distanceFromObject - 20);
  }

  return path;
}

Path arrowPointingRight(Path path, double contentWidth, double contentHeight,
    bool fitBellow, double distanceFromObject) {
  if (fitBellow) {
    path.relativeQuadraticBezierTo(-30, -30, 0, -distanceFromObject + 20);
  } else {
    path.relativeQuadraticBezierTo(-30, 30, 0, distanceFromObject - 20);
  }

  return path;
}

Path straightArrow(Path path, double contentWidth, double contentHeight,
    bool fitBellow, double distanceFromObject) {
  if (fitBellow) {
    path.relativeQuadraticBezierTo(0, 0, 0, -distanceFromObject + 20);
  } else {
    path.relativeQuadraticBezierTo(0, 0, 0, distanceFromObject - 20);
  }

  return path;
}
